/*
			 Salvatore Ruggieri (c), 2008
*/

/** @file basic.h  */

#ifndef _BASIC_H__INCLUDED_
#define _BASIC_H__INCLUDED_

#include "util.h"

/** Syntax for types, type declarations, modes and programs. */
namespace syn
{

/** Types. Syntax:
  * ! for definitiness,
  * |n for lower and upper bounds with range at most n,
  * |e for parametric typings,
  * | for lower and upper bounds,
  * _ for lower bound,
  * ^ for upper bound,
  * * for star. */
typedef enum {DEF, BOX, UP, DOWN, STAR} TypeValues;

/** %Type element. */
class Type
{
public:
	/** Type value */
	TypeValues tp;
	/** For |n types only, it is the n. */
	double range;
	/** For |e types only, it is the e. */
	string parametric;
	/** For tp==DEF only, true for parametric types. */
	bool is_p;

	/** Constructor from type values. */
	Type(TypeValues tval) {
		tp = tval;
		range = 0;
		is_p = false;
	}

	/** Constructor for type |n. */
	Type(double range) {
		if( range < 0 )
			throw runtime_error("Syntax error in type |r, r = " + string_util::tostr(range));
		tp = DEF;
		this->range = range;
		is_p = false;
	}

	/** Empty constructor, * type. */
	Type() {
		tp = STAR;
		range = 0;
		is_p = false;
	}
	
	/** Destructor. */
	virtual ~Type()
	{
	}

	/** Assignment operator from type values. */
	virtual Type& operator=(TypeValues other) {
		tp = other;
		range = 0;
		is_p = false;
		return *this;
	}
	
	/** Equality operator for type values. */
	virtual bool operator==(TypeValues other) const {
		return (tp == other) && (range == 0) && !is_p;
	}

	/** Equality operator. For parametric types, just check both are parametric.*/
	virtual bool operator==(const Type &other) const {
		if( tp != other.tp )
			return false;
		if( tp == DEF )
			if( is_p )
				return other.is_p && (parametric == other.parametric);
			else
				return (range == other.range) && (!other.is_p);
		return true;
	}
	
	/** Inequality operator. */
	virtual bool operator!=(const Type &other) const {
		return !(this->operator==(other));
	}
	
	/** Greater-than operator. */
	virtual bool operator>=(const Type &other) const {
		if( tp != DEF ) {
			if( tp == BOX )
				return other.tp >= BOX;
			else
				return (tp == other.tp) || (other.tp == STAR);
		}
		if( other.tp == DEF )
			if( is_p || other.is_p )
				return (is_p == other.is_p) && (parametric == other.parametric);
			else
				return range <= other.range;
		return true;
	}

	/** Greatest lower bound of two types. */
	static Type glb(const Type &v1, const Type &v2)
	{
		if( v1.is_p || v2.is_p )
			if( v1 >= v2 )
				return v1;
			else if( v2 >= v1 )
				return v2;
			else
				throw runtime_error("glb between parametric types: " + v1.to_string() + " and " + v2.to_string());
		if( v1 >= v2 ) return v1;
		if( v2 >= v1 ) return v2;
		return BOX;
	}

	/** Return the string representation of a type. */
	string to_string() const
	{
		switch( tp ) {
			case DEF:
				if( is_p )
					return "|"+parametric;
				if( range == 0 )
					return "!";
				return "|"+string_util::tostr(range);
				break;
			case BOX:
				return "|";
				break;
			case UP:
				return "^";
				break;
			case DOWN:
				return "_";
				break;
			case STAR:
				return "*";
				break;
			default:
				return "?";
				break;
		}
		return "?";
	}


	/** Parse a char to get the type. */
	Type(const string &s)
	{
		size_t len = s.length();
		if( len == 0 )
			throw runtime_error("Syntax error in typing: " + s);
		char c = s.at(0);
		range = 0;
		is_p = false;
		switch( c ) {
			case '!':
				tp = DEF;
				break;
			case '|':
				if( len == 1 )
					tp = BOX;
				else {
					tp = DEF;
					char *end;
					range = strtod(s.c_str()+1,&end);
					if( *end || (range < 0) ) {
						parametric = s.c_str()+1;
						range = 0;
						is_p = true;
					}
				}
				break;
			case '^':
				tp = UP;
				break;
			case '_':
				tp = DOWN;
				break;
			case '*':
				tp = STAR;
				break;
			default:
				throw runtime_error("Syntax error in typing: " + s);
				break;
		}
	}
	
	typedef enum {UP2BOX, UP2DEF, UP2BT, UP2PT} TypeSystem;

	/** Return which type system is used.
	  * UP2BOX if variables are typed in {*,_,^,|}
	  * UP2DEF if variables are typed in {*,_,^,|,!}
	  * UP2BT if variables are typed in {*,_,^,|,!,|r} for some r > 0
	  * UP2PT if parametric types
	  */
	TypeSystem typeSystem() const
	{
		if( tp == DEF ) {
			if( is_p )
				return UP2PT;
			if( range > 0 )
				return UP2BT;
			return UP2DEF;
		}
		return UP2BOX;
	}

};


/** %Type declaration. Stored in almost normal form, i.e., there is no duplicate 
   * type for a single variable, but atd x:* may appear. */
class TypeDecl : nocopy
{
public:

	/** Constructor. */
	TypeDecl(const string &s = "")
	{
		add(s);
	}

	/** Add type declaration. */
	void add(const string &s)
	{
//		cout << s << endl;
//		cout.flush();
		tokenizer tk(s.c_str(), ',');
		while( tk.has_next() )
			add_single(tk.next());
	}

	/** Length of the type declaration. */
	size_t size() const
	{
		return varname.size();
	}

	/** Find whether a variable is typed. */
	bool find(const string &var, size_t &pos) const
	{
		vector<string>::const_iterator it = ::find(varname.begin(), varname.end(), var);
		if( it == varname.end() )
			return false;
		pos = (size_t)(it - varname.begin());
		return true;
	}

	/** Get variable name. */
	const string& getVarName(size_t i) const
	{
		return varname[i];
	}

	/** Get variable type. */
	const Type& getType(size_t i) const
	{
		return vartype[i];
	}

	/** Return which type system is used.
	  * UP2BOX if variables are typed in {*,_,^,|}
	  * UP2DEF if variables are typed in {*,_,^,|,!}
	  * UP2BT if variables are typed in {*,_,^,|,!,|r} for some r > 0
	  * UP2PT if parametric types
	  */
	Type::TypeSystem typeSystem() const
	{
		Type::TypeSystem res = Type::UP2BOX;
		size_t sz = varname.size();
		for(size_t i=0; i<sz; ++i) {
			const Type &type = vartype[i];
			Type::TypeSystem t = type.typeSystem();
			if( t > res )  {
				if( t == Type::UP2PT )
					return t;
				res = t;
			}
		}
		return res;
	}

	/** Set variable type. */
	void setType(size_t i, Type v)
	{
		vartype[i] = v;
	}

	/** Set all variable to a given type. */
	void setAllType(Type v)
	{
		size_t sz = varname.size();
		for(size_t i=0; i<sz; ++i)
			vartype[i] = v;
	}

	/** Return a new type declaration obtained by restricting to a list of variables. */
	TypeDecl *restrict(const string &varlist) const
	{
		TypeDecl *c = new TypeDecl();
		tokenizer tk(varlist.c_str(), ',');
		size_t pos = 0;
		while( tk.has_next() ) {
			const char *var = tk.next();
			// in case of duplicates in the varlist
			if( c->find(var, pos) )
				continue;
			c->varname.push_back(var);
			if( find(var, pos)  )
				c->vartype.push_back(vartype[pos]);
			else
				c->vartype.push_back(STAR);
		}
		return c;
	}

	/** Return true if the type declaration implies a given one. */
	bool implies(const TypeDecl *d) const
	{
		size_t nvar = d->varname.size();
		size_t pos;
		for(size_t i=0;i<nvar;++i) {
			Type tp = d->vartype[i];
			if( tp == STAR )
				continue;
			if( !find( d->varname[i], pos ) ) 
				return false;
			Type tp2 = vartype[pos];
			if( !(tp2 >= tp) )
				return false;
		}
		return true;
	}

	/** Get the list of typed variables. */
	string getVarList() const
	{
		string res;
		size_t nvar = varname.size();
		for(size_t i=0;i<nvar;++i) {
			res.append(varname[i]);
			if( i != nvar-1 )
				res.append(",");
		}
		return res;
	}

	/** Get string representation. */
	string to_string() const
	{
		size_t nvar = varname.size();
		if( nvar == 0 )
			return "";
		string res;
		size_t end = nvar - 1;
		for(size_t i=0;i<nvar;++i) {
			res.append(varname[i]);
			res.append(":");
			res.append(vartype[i].to_string());
			if( i != end )
				res.append(",");
		}
		return res;
	}

	/** Stream output. */
	friend ostream& operator <<(ostream &os, const TypeDecl &decl)
	{
		os << decl.to_string();
		return os;
	}

private:
	/** Variable types. */
	vector<Type> vartype;
	/** Variable names. */
	vector<string> varname;

	void add_single(const string &s)
	{
		size_t loc = s.find(':');
		if( loc == string::npos )
			throw runtime_error("Parse error: " + string(s));
		char *type = string_util::trim(s.substr(loc+1).c_str());
		Type tp( type );
		// search for duplicates
		char *var = string_util::trim(s.substr(0, loc).c_str());
		vector<string>::iterator it = ::find(varname.begin(), varname.end(), var);
		if( it == varname.end() ) {
			varname.push_back( var );
			vartype.push_back(tp);
		}
		else {
			int pos = it - varname.begin();
			vartype[pos] = Type::glb(vartype[pos], tp);
		}
		delete []type;
		delete []var;
	}
};


/** Methods for parsing terms and atoms with callback actions. */
template<class F>
class Parsing 
{
public:
	/** Constructor. */
	Parsing(F &a)
	: action(a)
	{
	}
	
	/** Set error description. */
	void set_err(const char *err)
	{
		err_desc = err;
	}
	
	/** Parse an atom., */
	const char *parse_atom(const char *head) 
	{
		string res;
		const char *tail = parse_id(head);
		action.atom( string(head, tail-head) );
		head = tail;
		if( *head == '(' ) {
			do {
				const char *tail = parse_term(++head);
				action.atom_arg( string(head, tail-head) );
				head = tail;
			} while( *head == ',' );
			if( *head != ')' )
				throw runtime_error("Unknown syntax: " + string(head));
			++head;
		}
		return head;
	}
	
	/** Parse terms. Terms are defined by the following grammatic:
		T ::= a | V | f(LT) | TL 		-- term
		TL ::= [] | [LT] | [LT|TL] 	-- term list
		LT ::= T | T,LT			-- list of term
		V ::= _ | Uppercaseidentifier	-- variable
	*/
	const char *parse_term(const char *head) 
	{
		// looks for lists
		if( *head == '[' )
			return parse_termlist(head);
		// variable
		if( (*head <= 90) && (*head >= 65) )
			return parse_var(head);
		// functor or constant
		head = parse_id(head);
		if( *head == '(' ) {
			// functor
			head = parse_list(++head);
			if( *head != ')' )
				throw runtime_error("Unknown syntax: " + string(head) + " in: " + string(err_desc));
			return ++head;
		}
		// constant
		return head;
	}

	/** Parse a term which is a list, */
	const char *parse_termlist(const char *head) 
	{
		// looks for lists
		if( *head == '[' ) {
			++head;
			if( *head == ']' ) 
				return ++head;
			head = parse_list(head);
			if( *head == ']' )
				return ++head;
			if( *head == '|' ) {
				++head;
				if( *head == '[' ) 
					head = parse_termlist(head);
				else
					head = parse_var(head);
				if( *head != ']' )
					throw runtime_error("Unknown syntax: " + string(head) + " in: " + string(err_desc));
				return ++head;
			}
		}
		throw runtime_error("Unknown syntax: " + string(head) + " in: " + string(err_desc));
	}
	
	/** Parse a variable. */
	const char *parse_var(const char *head) 
	{
		// looks for variable _
		if( *head == '_' ) {
			action.var( "_" + string_util::tostr(unique::id++) );
			return ++head;
		}
		// any other variable starts uppercase
		if( (*head > 90) || (*head < 65) )
				throw runtime_error("Unknown syntax: " + string(head) + " in: " + string(err_desc));
		const char *tail = parse_id(head+1);
		action.var( string(head, (tail-head)) );
		return tail;
	}

	/** Parse a list of terms. */
	const char *parse_list(const char *head) 
	{
		--head;
		do {
			++head;
			head = parse_term(head);
		} while( *head == ',' );
		return head;
	}
	
	/** Parse an identifier (variable or functor)., */
	const char *parse_id(const char *head)
	{
		while( *head )
			if( (*head < 48) || (*head > 57 && *head < 65) || (*head > 90 && *head < 95) || (*head> 122) )
				break;
			else
				++head;
		return head;
	}

private:
	/** Callback action. */
	F &action;
	/** Pointer to error description. */
	const char *err_desc;
};

/** Callback interface for the %Parsing class. */
class ParseAction
{
public:
	/** Parsing a variable. */
	void var(const string &name) {}
	/** Parsing an atom. */
	void atom(const string &name) {}
	/** Parsing an atom argument. */
	void atom_arg(const string &name) {}
};

/** Callback class collecting the list of variables typed by a provided type. */
class ParseVar : public ParseAction
{
public:
	/** Result. */
	string res;
	/** Type. */
	string type;
	/** Constructor. */
	ParseVar(const string &tp)
	: type(tp)
	{
	}
	void var(const string &name) 
	{
		if( res != "" )
			res += ",";
		res += name + ":" + type;
	}
};

/** Callback class collecting the atom predicate and the list of its arguments. */
class ParseAtom : public ParseAction
{
public:
	/** Predicate identifier. */
	string predicate;
	/** List of arguments. */
	string argslist;
	/** No of arguments. */
	size_t argsize;
	
	ParseAtom() 
	: argsize(0) 
	{
	}
	void atom(const string &name) 
	{
		predicate = name;
	}
	void atom_arg(const string &name) 
	{
		if( argsize > 0 )
			argslist += '\n';
		++argsize;
		argslist += name;
	}
};

/** %Mode for a predicate. */
class Mode : nocopy
{
public:

	/** Constructor. */
	Mode(const string &atom = "")
	: predarity(0), in(NULL), out(NULL)
	{
		set(atom);
	}

	/** Destructor. */
	virtual ~Mode()
	{
		if( in != NULL )
			delete[] in;
		if( out != NULL )
			delete[] out;
	}

	/** Set mode. */
	void set(const string &atom)
	{
		// BUG: improve the code of this method
		char *_atm = string_util::remove_spaces(atom.c_str());
		char *end_token = (char *)string_util::find_sep(_atm, '(');
		if( *end_token == 0 ) {
			predname = _atm;
			predarity = 0;
			in = out = NULL;
			delete[] _atm;
			return;
		}
		predname = string(_atm, end_token - _atm);
		++end_token;
		char *last = end_token + strlen(end_token) - 1;
		if( *last != ')' )
			throw new runtime_error("Syntax error: " + atom);
		*last = 0;
		tokenizer tk( end_token, ',');
		predarity = tk.count();
		in = new Type[predarity];
		out = new Type[predarity];
		if( tk.count() != predarity )
			throw new runtime_error("Syntax error: " + atom);
//		cout << "Processing: " << atom << endl;
		for(size_t i=0; i<predarity; ++i) {
			const char *iotype = tk.next();
			char *x = (char *)string_util::find_sep(iotype, 'x');
			if( *x == 0 )
				throw new runtime_error("Syntax error: " + atom);
			string intype = string(iotype, x - iotype);
			in[i] = Type(intype);
			out[i] = Type(x+1);
		}
		delete[] _atm;
//		cout << "End Processing" << endl;
	}

	/** Get predicate name. */
	const string &getPredicate() const
	{
		return predname;
	}

	/** Get predicate arity. */
	size_t getArity() const
	{
		return predarity;
	}

	/** Get input modes for given array of variables. */
	string getInputModes(const string *vars) const
	{
		return _get(vars, in);
	}

	/** Get output modes for given array of variables. */
	string getOutputModes(const string *vars) const
	{
		return _get(vars, out);
	}

	/** Return which type system is used.	  */
	Type::TypeSystem typeSystem() const
	{
		Type::TypeSystem res = Type::UP2BOX;
		for(size_t i=0; i<predarity; ++i) {
			Type::TypeSystem t = in[i].typeSystem();
			if( t > res )  {
				if( t == Type::UP2PT )
					return t;
				res = t;
			}
		}
		for(size_t i=0; i<predarity; ++i) {
			Type::TypeSystem t = out[i].typeSystem();
			if( t > res )  {
				if( t == Type::UP2PT )
					return t;
				res = t;
			}
		}
		return res;
	}

	/** Get string representation. */
	string to_string() const
	{
		string s = predname;
		if( predarity == 0 )
			return s;
		s.append("(");
		size_t end = predarity-1;
		for(size_t i=0; i<=end; ++i) {
			s.append(in[i].to_string());
			s.append("x");
			s.append(out[i].to_string());
			if( i < end )
				s.append(",");
		}
		s.append(")");
		return s;
	}

	/** Output to a stream. */
	friend ostream& operator <<(ostream &o, const Mode &m)
	{
		o << m.to_string();
		return o;
	}

private:
	/** Predicate name. */
	string predname;
	/** Predicate arity. */
	size_t predarity;
	/** Input types. */
	Type *in;
	/** Output types. */
	Type *out;

	/** From predicate arguments to type declarations. */
	string _get(const string *vars, Type *modearray) const
	{
		string res;
		for(size_t i=0; i<predarity; ++i)
		{
			ParseVar action(modearray[i].to_string());
			Parsing<ParseVar> parse(action);
			const char *last = parse.parse_term(vars[i].c_str());
			if( *last != 0 )
				throw runtime_error("Unknown syntax: " + vars[i]);
			if( action.res == "" )
				continue;
			if( res != "" )
				res += ",";
			res += action.res;
		}
		return res;
	}
};


/** A moded atom. */
class Atom : nocopy
{
public:

	/** Constructor. */
	Atom(const string &atom = "")
	: predarity(0),vars(NULL)
	{
		set(atom);
	}

	/** Destructor. */
	virtual ~Atom()
	{
		if( vars != NULL )
			delete[] vars;
	}

	/** Set moded atom. */
	void set(const string &atom)
	{
		if( vars != NULL )
			delete[] vars;
		char *_atm = string_util::remove_spaces(atom.c_str());
		ParseAtom action;
		Parsing<ParseAtom> parser(action);
		parser.set_err(_atm);
		const char *head = parser.parse_atom(_atm);
		if( *head != 0 )
			throw new runtime_error("Syntax error: " + atom);
		predname = action.predicate;
		string argterms = action.argslist;
		predarity = action.argsize;
		if( predarity == 0) {
			vars = NULL;
			delete[] _atm;
			return;
		}
		tokenizer tk( argterms.c_str(), '\n');
		vars = new string[predarity];
		for(size_t i=0; i<predarity; ++i ) 
			vars[i] = tk.next();
		delete[] _atm;
	}

	/** Get predicate name. */
	const string &getPredicate() const
	{
		return predname;
	}

	/** Get predicate arity. */
	size_t getArity() const
	{
		return predarity;
	}

	/** Get predicate mode. */
	Mode *findMode(const vector<Mode *> &modes) const
	{
		// BUG: [efficiency] linear search is implemented so far
		size_t sz = modes.size();
		for(size_t i=0; i<sz; ++i ) {
			if( predarity == modes[i]->getArity() &&
				predname == modes[i]->getPredicate() )
				return modes[i];
		}
		return NULL;
	}

	/** Get input modes for variables. */
	string getInputModes(Mode *m) const
	{
		return m->getInputModes(vars);
	}

	/** Get output modes for variables. */
	string getOutputModes(Mode *m) const
	{
		return m->getOutputModes(vars);
	}

	/** Get modes for variables by typing as the passed type argument. */
	string getModes(const Type &tp) const
	{
		ParseVar action(tp.to_string());
		Parsing<ParseVar> parse(action);
		for(size_t i=0; i<predarity; ++i)
		{
			const char *last = parse.parse_term(vars[i].c_str());
			if( *last != 0 )
				throw runtime_error("Unknown syntax: " + vars[i]);
		}
		return action.res;
	}

	/** Get string representation. */
	string to_string() const
	{
		string s = predname;
		if( predarity == 0 )
			return s;
		s.append("(");
		size_t end = predarity-1;
		for(size_t i=0; i<end; ++i) {
			s.append(vars[i]);
			s.append(",");
		}
		s.append(vars[end]);
		s.append(")");
		return s;
	}

	/** Output to a stream. */
	friend ostream& operator <<(ostream &o, const Atom &m)
	{
		o << m.to_string();
		return o;
	}

private:
	/** Predicate name. */
	string predname;
	/** Predicate arity. */
	size_t predarity;
	/** Array of variables. */
	string *vars;
};


/** A program rule. */
class Rule : nocopy
{
public:

	/** Constructor. */
	Rule(const string &rule = "")
	{
		set(rule);
	}

	/** Destructor. */
	virtual ~Rule()
	{
		size_t len = body.size();
		for(size_t i=0; i<len; ++i)
			delete body[i];
	}

	/** Set the rule. */
	void set(const string &rule)
	{
		// BUG: improve the code of this method
		size_t len = body.size();
		for(size_t i=0; i<len; ++i)
			delete body[i];
		body.clear();
		constraint = "";

		size_t loc = rule.find(":-");
		if( loc == string::npos ) {
			// atomic rule
			head.set(rule);
			return;
		}
		head.set( rule.substr(0, loc) );
		loc++;
		size_t loc1 = rule.find('{');
		if( loc1 != string::npos ) {
			size_t loc2 = rule.find('}');
			if( loc2 == string::npos )
				throw new runtime_error("Syntax error: " + rule);
			char *realc = string_util::remove_spaces(rule.c_str()+loc1+1, rule.c_str()+loc2);
			constraint = realc;
			delete[] realc;
			loc = loc2;
		}
		char *realbody = string_util::remove_spaces(rule.c_str()+loc+1);
		// BUG: inline the _parse_body method
		string res = _parse_body(realbody);
		delete []realbody;
		tokenizer tk(res.c_str(), '\n');
		while( tk.has_next() )
			body.push_back(new Atom(tk.next()));
	}

	/** Get rule head atom. */
	const Atom *getHead() const
	{
		return &head;
	}

	/** Get rule constraint. */
	const string &getConstraint() const
	{
		return constraint;
	}

	/** Get number of atoms in the body. */
	size_t size() const
	{
		return body.size();
	}

	/** Get body atom. */
	const Atom *getBodyAtom(size_t i) const
	{
		return body[i];
	}

	/** Return which type system is used.	  */
	Type::TypeSystem typeSystem(const vector<Mode *> &modes) const
	{
		Type::TypeSystem res = head.findMode(modes)->typeSystem();
		size_t sz = body.size();
		for(size_t i=0; i<sz; ++i) {
			Type::TypeSystem t = body[i]->findMode(modes)->typeSystem();
			if( t > res )  {
				if( t == Type::UP2PT )
					return t;
				res = t;
			}
		}
		return res;
	}

	/** Get string representation. */
	string to_string() const
	{
		string s = head.to_string();
		size_t len = body.size() ;
		if( constraint != "" ) {
			s.append(" :- {");
			s.append(constraint);
			s.append("}");
			if( len > 0 )
				s.append(",");
		} else if( len > 0 )
			s.append(" :- ");
		if( len > 0 ) {
			size_t end = len-1;
			for(size_t i=0; i<end; ++i) {
				s.append( body[i]->to_string() );
				s.append(",");
			}
			s.append( body[end]->to_string() );
		}
		s.append(".");
		return s;
	}

	/** Output to a stream. */
	friend ostream& operator <<(ostream &o, const Rule &m)
	{
		o << m.to_string();
		return o;
	}

private:
	/** Head. */
	Atom head;
	/** Constraint. */
	string constraint;
	/** Array of body atoms. */
	vector<Atom *> body;

	static string _parse_body(const char *head) 
	{
		if( *head == 0 )
			return "";
		string res;
		--head;
		ParseAction no;
		Parsing<ParseAction> parser(no);
		do {
			++head;
			parser.set_err(head);
			const char *tail = parser.parse_term(head);
			if( res != "" )
				res += "\n";
			res += string(head, tail-head);
			head = tail;
		} while( *head == ',' );
		if( *head != 0 )
			throw runtime_error("Unknown syntax: " + string(head));
		return res;
	}
};


/** A moded set of rules. */
class Program : nocopy
{
public:

	/** Constructor. */
	Program(const string &program = "")
	{
		set(program);
	}

	/** Destructor. */
	virtual ~Program()
	{
		clear();
	}

	/** Set program from source code. */	/** Parse a list of terms., */
	/** Parse a list of terms., */

	void set(const string &program)
	{
		clear();

		const char *s = program.c_str();
		while( *s )
		{
			//  skip break chars
			if( *s == ' ' || *s == '\n' || *s == '\t' || *s == '\r' ) {
				++s;
				continue;
			}
			// either a rule or a mode annotation
			const char *p = string_util::find_sep2(s, '-', '.');
			if( !*p ) throw new runtime_error("Syntax error: " + program);
			// either a fact or a mode annotation
			if( *p == '.' ) {
				string r(s, 0, p-s);
				if( strncmp(r.c_str(), "%mode ", 6) == 0 )
					modes.push_back( new Mode( string(r, 6) ) );
				else
					rules.push_back( new Rule( r ) );
				s = p+1;
				continue;
			}
			// a non-fact rule
			p = string_util::find_sep2(p+1, '{', '.');
			if( !*p ) throw new runtime_error("Syntax error: " + program);
			if( *p == '{' ) {
				p = string_util::find_sep(p+1, '}');
				if( !*p ) throw new runtime_error("Syntax error: " + program);
				p = string_util::find_sep(p+1, '.');
				if( !*p ) throw new runtime_error("Syntax error: " + program);
			}
			rules.push_back( new Rule( string(s, 0, p-s) ) );
			s = p+1;
		}

	}

	/** Output to a stream. */
	friend ostream& operator <<(ostream &o, Program &r)
	{
		size_t len = r.modes.size();
		for(size_t i=0; i<len; ++i)
			o << "%mode " << *(r.modes[i]) << "." << endl;
		len = r.rules.size();
		for(size_t i=0; i<len; ++i)
			o << *(r.rules[i]) << endl;
		return o;

	}
	
	/** Parse an input stream containing a program. */
	static string parse(istream &in)
	{
		string rules;
		rules.reserve(16*1024);
		while( !in.eof() ) {
			string l;
			getline(cin, l);
			size_t sz = l.size();
			// remove trail \r (Windows editors)
			if( sz > 0 && l[sz-1] == '\r' )
				l.resize(sz-1);
			if( l == "" )
				continue;
			const char *s = l.c_str();
			// rules out comments (except mode annotations)
			if( *s == '%' && strncmp(s, "%mode", 5) )
				continue;
			rules.append(l);
			rules.append("\n");
		}
		return rules;
	}
	
	/** Return which type system is used.	  */
	Type::TypeSystem typeSystem() const
	{
		Type::TypeSystem res = Type::UP2BOX;
		size_t nmodes = modes.size();
		for(size_t i=0; i<nmodes; ++i) {
			Type::TypeSystem t = modes[i]->typeSystem();
			if( t > res )  {
				if( t == Type::UP2PT )
					return t;
				res = t;
			}
		}
		return res;
	}

	/** Predicate modes. */
	vector<Mode *> modes;
	/** Program rules. */
	vector<Rule *> rules;

private:
	/** Clear data. */
	void clear()
	{
		size_t n = modes.size();
		for(size_t i=0; i<n; ++i)
			if( modes[i] != NULL )
				delete modes[i];
		n = rules.size();
		for(size_t i=0; i<n; ++i)
			if( rules[i] != NULL )
				delete rules[i];
	}
};



} // end of namespace clptsyn

using namespace syn;

#endif // !defined(_BASIC_H__INCLUDED_)

